//
// A class to manage ODF constituent information.
//
// 1998, Giuseppe Vacanti (XMM-SOC ESA)
//
#include <string>    // C++

#include "Dal.h"     // SAS
#include "errstr.h"
#include "Tracer.h"
#include <stdlib.h>

#include "OdfConstituent.h" // this application

OdfConstituent::OdfConstituent() {
 PROC
 ENDPROC
}

OdfConstituent::OdfConstituent(string fname) :
  _name(fname) {

  PROC

  if(!atoi(revolution().c_str()))
    errstr << "Prospective ODF constituent: " << name() << ". The revolution identifier " << revolution() << " is not an integer > 0. Rename this file to comply with the ODF ICD." << Sas::error("InvalidRevolutionIdentifier"); 
  ENDPROC

}

// The following member functions return substrings from
// the name of an ODF component. 
//
// RRRR_PPPPPPOOLL_IIUEEECCMMF.ZZZ
// 0123456789012345678901234567890
// 0         1         2         3
//
// As per ODF ICD Issue 2.0:
//  - RRRR         revolution number (integer)
//  - PPPPPP       proposal identifier. First P=9 => slew
//  - OOLL         observation identifier
//  - II           OM R1 R2 M1 M2 PN RM SC
//  - U            Scheduled or Unscheduled or Not applicable (X)
//  - EEE          exposure identifier. 000 => file not related to exposure periods.
//  - CC           CCD identifier, OM window identifier
//  - MM           IM TI RI CT AU ... BU ... SP HT ... FA ... AT
//  - F            E I X H S M
//
//
// We manipulate everything as strings. This is enough for the OAL.
//
string OdfConstituent::name() const {
  PROC
  RETURN _name;
}

string OdfConstituent::revolution() const {
  PROC
  RETURN this->base_name().substr(0,4);
}

string OdfConstituent::observation() const {
  PROC
  RETURN this->base_name().substr(5,10);
}

string OdfConstituent::proposal() const {
  PROC
  RETURN this->base_name().substr(5,10);
}

string OdfConstituent::proposal_exposure() const {
  PROC
  RETURN this->base_name().substr(5,10) + this->instrument() +
    this->exposure();
}

string OdfConstituent::instrument() const {
  PROC
  RETURN this->base_name().substr(16,2);
}

string OdfConstituent::scheduled() const {
  PROC
  RETURN this->base_name().substr(18,1);
}

string OdfConstituent::sux_exposure() const {
  PROC
  RETURN scheduled() + exposure();
}

string OdfConstituent::exposure() const {
  PROC
  RETURN this->base_name().substr(19,3);
}

// These work because we know that file names are 31 characters
// long. Ideally one would have to scan the string from the end and
// look for a /. At the moment these methods don't work well with g++.
string OdfConstituent::base_name() const {
  PROC
    RETURN this->name().substr(this->name().length() - 31, 31);
}

string OdfConstituent::dir_name() const {
  PROC
  RETURN name().substr(0,name().length()-31);
}
    

string OdfConstituent::ccd() const {
  PROC
  RETURN this->base_name().substr(22,2);
}

string OdfConstituent::data() const {
  PROC
  RETURN this->base_name().substr(24,2);
}

string OdfConstituent::type() const {
  PROC
  RETURN this->base_name().substr(26,1);
}

bool OdfConstituent::is_slew() const {
  PROC
  if(this->proposal().substr(0,1) == "9") ALTRETURNV(true)
  RETURN false;
}

FileType OdfConstituent::file_type() const {
  PROC
  string sft = this->type();
  FileType ft;

  if(sft == "E" || sft == "I")
    ft = science;
  else if(sft == "X")
    ft = auxiliary;
  else if(sft == "H") 
    ft = housekeeping;
  else if(sft == "S")
    ft = spacecraft;
  else if(sft == "M")
    ft = summary;
  else
    errstr << this->name() << Sas::error("UnknowOdfFileType");
  
  RETURN ft;
}
    
