//
// Odf.cc
//
// Giuseppe Vacanti, XMM-SOC, European Space Agency
// 1998-1999
//
//  The main method is scan(). This finds all correctly named ODF
//  components in directory. Then each component is examined and its
//  main parameters (instrument, date start, date end) extracted and
//  used to build the ODF summary file.
//
//  file_set -- set tallying the names of the dataset that have
//  already been processed. Can be used to prevent processing of some
//  datasets. 
//
//  file_list -- these datasets will be listed in the ODF summary
//  file. Can be used to add datasets that are either not readable
//  with the DAL, or have missing components (date-xxxx typically).

#include <string>      // C++
#include <vector>
#include <set>
#include <multimap.h>
#include <iostream.h>
#include <set>
#include <fstream.h>
#include <iomanip.h>

#include <stdlib.h>

#include <Tracer.h>    // SAS
#include <metatask.h>
#include <errstr.h>
#include <Dal.h>
#include <param.h>
#include <Xmm.h>
#include <STime.h>
#include <stringutil.h>
#include <SasDalUtils.h>
#include <OalServer.h>

#include "ExposureObject.h"
#include "OdfObject.h"
#include "OdfConstituent.h" // this application

#include "ahfgen.h"
#include "phkgen.h"
#include "nphkgen.h"
#include "thfgen.h"
#include "wdgen.h"
#include "omphkgen.h"
#include "tcfgen.h"
#include "rgsauxtable.h"

// This function is made redundant by STime's new validation methods.
// s ~ yyyy-mm-ddThh:mm:ss[.sss...]?
static bool isValidFitsTimeString(const string s)
{
  PROC
  vector<string> date_time = breakup(s, "T");
  if(date_time.size() != 2){
    // cerr << "fail1" << endl;
    ALTRETURNV(false);
  }
  vector<string> date = breakup(date_time[0], "-");
  if(date.size() != 3) {
    // cerr << "fail2" << endl;
    ALTRETURNV(false);
  } else {
    if(date[0].size() != 4){
      // cerr << "fail2" << endl;
      ALTRETURNV(false);
    }
    if(date[1].size() != 2 || atoi(date[1].c_str()) <= 0 || atoi(date[1].c_str()) > 12){
      // cerr << "fail4" << endl;
      ALTRETURNV(false);
    }
    if(date[2].size() != 2 || atoi(date[2].c_str()) <= 0 || atoi(date[2].c_str()) > 31){ // Should really do a month-dependent check here
      // cerr << "fail5" << endl;
      ALTRETURNV(false);
    }
  }
  vector<string> time = breakup(date_time[1], ":");
  if(time.size() != 3) {
    // cerr << "fail6" << endl;
    ALTRETURNV(false);
  } else {
    if(time[0].size() != 2 || atoi(time[0].c_str()) < 0 || atoi(time[0].c_str()) > 24){
      // cerr << "fail7" << endl;
      ALTRETURNV(false);
    }
    if(time[1].size() != 2 || atoi(time[1].c_str()) < 0 || atoi(time[1].c_str()) > 59){
      // cerr << "fail8" << endl;
      ALTRETURNV(false);
    }
    if(time[2].size() < 2 || atof(time[2].c_str()) < 0 || atof(time[2].c_str()) > 60){ // this covers the leap second
      // cerr << "fail9" << endl;
      ALTRETURNV(false);
    }
  }
  ALTRETURNV(true);  
}


OdfObject::OdfObject(string dir)
  : _dir(dir),
    _pointing_from_ats(true),
    _ra(0.),
    _dec(0.),
    _posangle(90.),
    _has_good_duration(false),
    _has_good_start(false),
    _has_good_end(false)
{
  PROC
  init(); 

  // This constructor is called when we want to extract the pointing
  // information from an already existing ATS (attitude history
  // file). init() performs a directory scan, and at this stage we
  // know if the ATS exists or not.

  if(dataSetServer->exists(_dir + "/" + ahf())) {
    DataSet * set = dataSetServer->open(_dir + "/" + ahf(), Dal::Read,
					DataSetServer::Low);
    Table * tab = set->table("SCATS1");

    // According to XMM-MOC-ICD-0006-OAD the first record in the ATS
    // is the nominal pointing for the observation.
    //
    // VIEWRA                     11A       (11)      HH:MM:SS.SS
    // VIEWDECL                   11A       (11)      sDD:MM:SS.S
    // ASTPOS                     1E        (1)       DEGREE

    using namespace SasDalUtils;
    ColumnData * ptr;
    ptr = tab->column("VIEWRA")->data();
    vector<string> v = breakup(getStringColumnRow(ptr),":");
    _ra = (atof(v[0].c_str()) + atof(v[1].c_str())/60. + atof(v[2].c_str())/3600.)*15.;
    ptr = tab->column("VIEWDECL")->data();
    v = breakup(getStringColumnRow(ptr),":");
    _dec = (atof(v[0].c_str()) + atof(v[1].c_str())/60. + atof(v[2].c_str())/3600.);

    ptr = tab->column("ASTPOS")->data();
    _posangle = ptr->real32Data()[0];
    delete ptr;
    errstr << "Using the nominal pointing direction from " + _dir +
      "/" + ahf() + " : RA = " << _ra << " DEC = " << _dec << " PA = " << _posangle << Sas::message(Msg::App, Msg::Sparse);
    dataSetServer->close(set);
  }
  else
    errstr << "I was asked to extract the pointing information from the attitude history file, but no such file exists. Using default values: RA = " << _ra << " DEC = " << _dec << " PA = " << _posangle << Sas::warning("DefaultPointing");
  ENDPROC
}

OdfObject::OdfObject(double ra, double dec, double posangle, string dir)
  : _dir(dir),
    _pointing_from_ats(false),
    _ra(ra),
    _dec(dec),
    _posangle(posangle)
{
  PROC
  init();
  ENDPROC
}

void OdfObject::init()
{
  PROC
  instruments.push_back("OM");
  long_names["OM"] = "OM"; 
  fits_names["OM"] = "OM";

  instruments.push_back("M1");
  long_names["M1"] = "EPIC MOS";
  fits_names["M1"] = "EMOS1";

  instruments.push_back("M2");
  long_names["M2"] = "EPIC MOS";
  fits_names["M2"] = "EMOS2";

  instruments.push_back("R1");
  long_names["R1"] = "RGS";
  fits_names["R1"] = "RGS1";

  instruments.push_back("R2");
  long_names["R2"] = "RGS";
  fits_names["R2"] = "RGS2";

  instruments.push_back("PN");
  long_names["PN"] = "EPIC PN";
  fits_names["PN"] = "EPN";

  start_time = STime::latest;
  end_time = STime::earliest;

  // On the first scan, check the dates in the files.
  scan(true);
  // validate only the first time we scan the directory, when we have
  // not yet added any of our own ODF files. 
  validate();
  report();
  ENDPROC
}  

STime OdfObject::start() const {
  PROC
  RETURN start_time;
}

STime OdfObject::stop() const {
  PROC
  RETURN end_time;
}

unsigned int OdfObject::numberOfFiles() const {
  PROC
  RETURN file_list.size();;
}

void OdfObject::generateAuxiliaryFiles() {
  PROC

  set<string,less<string> > new_files;
  string new_file;
  pair<set<string, less<string> >::const_iterator, bool> p;

  // we only do this for OM, and eventually we hope not to have to
  // do this at all
  string s = "OM";
  if(is_instrument_active(s)) {
    typedef multimap<string,ExposureObject,less<string> >::const_iterator EI;
    EI lb = inst_exp_time_map.lower_bound(s);
    EI ub = inst_exp_time_map.upper_bound(s);
    for(EI i = lb; i != ub; i++) {
      const ExposureObject &e = (*i).second;
      new_file = _dir + "/" + thdf(s, e.exposure);
      Directory *d = new Directory(_dir);
      d->filter(thdf(s, e.exposure));
      if(d->size() == 0){
	p = new_files.insert(new_file);
	if(p.second) {
	  openParameters("thfgen");
	  setParameter("exposure", e.stop - e.start);
	  setParameter("thf", new_file);
	  setParameter("tstart", e.start.toString());
	  setParameter("tstop", e.stop.toString());
	  setParameter("expid", e.exposure);
	  setParameter("obsid", observation_id);
	  errstr << "Generating THDF " + new_file <<
	    Sas::message(Msg::App,Msg::Sparse);
	  task(thfgen);
	  closeParameters();
	}
      } else {
	errstr << "The OM tracking history data file (" + thdf(s, e.exposure) + ") exists and will not be overwritten. If this is not what you want, remove " + _dir + "/" + thdf(s, e.exposure) + " and start again." << Sas::warning("OMTrackingHistoryDataFileExists");
      }
      new_file = _dir + "/" + pwdf(s, e.exposure);
      d = new Directory(_dir);
      d->filter(pwdf(s, e.exposure));
      if(d->size() == 0){
	p = new_files.insert(new_file);
	if(p.second) {
	  openParameters("wdgen");
	  setParameter("wdf", new_file);
	  errstr << "Generating PWDF " + new_file <<
	    Sas::message(Msg::App,Msg::Sparse);
	  task(wdgen);
	  closeParameters();
	}
      } else {
	errstr << "The OM priority window data file (" + pwdf(s, e.exposure) + ") exists and will not be overwritten. If this is not what you want, remove " + _dir + "/" + pwdf(s, e.exposure) + " and start again." << Sas::warning("OMPriorityWindowDataFileExists");
      }
    }
  }

  ENDPROC
    }

void OdfObject::generateHousekeepingFiles() {
  PROC

    set<string,less<string> > new_files;
  string new_file;
  pair<set<string, less<string> >::const_iterator, bool> p;

    // only PN has a main periodic hk file
    string s = "PN";
  if(is_instrument_active(s)) {
    typedef multimap<string,ExposureObject,less<string> >::const_iterator EI;
    EI lb = inst_exp_time_map.lower_bound(s);
    EI ub = inst_exp_time_map.upper_bound(s);
    for(EI i = lb; i != ub; i++) {
      const ExposureObject &e = (*i).second;
      new_file = _dir + "/" + mphf(s, e.exposure);
      p = new_files.insert(new_file);
      if(p.second) {
	openParameters("phkgen");
	setParameter("extname", "MAIN_PERIODIC_HK");
	setParameter("phkf", new_file);
	setParameter("tstart", e.start.toString());
	setParameter("tstop", e.stop.toString());
	setParameter("instrument", fits_names[s]);
	errstr << "Generating Main PHKF " + new_file <<
	  Sas::message(Msg::App,Msg::Sparse);
	errstr << "TSTART = " << e.start << " TSTOP  = " << e.stop << Sas::message(Msg::App,Msg::Sparse);
	task(phkgen);
	closeParameters();
      }

    }
  }

  // only do the OM non periodic hk file.
  s = "OM";
  if(is_instrument_active(s)) {
    typedef multimap<string,ExposureObject,less<string> >::const_iterator EI;
    EI lb = inst_exp_time_map.lower_bound(s);
    EI ub = inst_exp_time_map.upper_bound(s);
    for(EI i = lb; i != ub; i++) {
      const ExposureObject &e = (*i).second;
      new_file = _dir + "/" + nphf(s, e.exposure);
      p = new_files.insert(new_file);
      if(p.second) {
	openParameters("nphkgen");
	setParameter("nphkf", new_file);
	setParameter("tstart", e.start.toString());
	setParameter("tstop", e.stop.toString());
	setParameter("instrument", fits_names[s]);
	errstr << "Generating  NPHKF " + new_file <<
	  Sas::message(Msg::App,Msg::Sparse);
	errstr << "TSTART = " << e.start << " TSTOP  = " << e.stop << Sas::message(Msg::App,Msg::Sparse);
	task(nphkgen);
	closeParameters();
      }

    }
  }
    

  // periodic housekeeping -- every instrument except PN got one
  typedef vector<string>::const_iterator VI;
  typedef multimap<string,ExposureObject,less<string> >::const_iterator EI;
  for(VI i = instruments.begin(); i != instruments.end(); i++) {

    if((*i) == "PN")
      continue;
    
    if(is_instrument_active(*i)) {
      EI lb = inst_exp_time_map.lower_bound(*i);
      EI ub = inst_exp_time_map.upper_bound(*i);
      for(EI j = lb; j != ub; j++) {
	const ExposureObject &e = (*j).second;
	if((*i) == "OM") { // ... but OM is special
	  new_file = _dir + "/" + phf(*i, e.exposure);
	  p = new_files.insert(new_file);
	  if(p.second) {
	    openParameters("omphkgen");
	    setParameter("phkf", new_file);
	    errstr << "Generating (OM) PHKF " + new_file <<
	      Sas::message(Msg::App,Msg::Sparse);
	    task(omphkgen);
	    closeParameters();
	  }
	} else {
	  new_file = _dir + "/" + phf(*i, e.exposure);
	  p = new_files.insert(new_file);
	  if(p.second) {
	    openParameters("phkgen");
	    setParameter("phkf", new_file);
	    setParameter("tstart", start().toString());
	    setParameter("tstop", stop().toString());
	    setParameter("instrument", fits_names[*i]);
	    errstr << "Generating PHKF " + new_file <<
	      Sas::message(Msg::App,Msg::Sparse);
	    errstr << "TSTART = " << start() << " TSTOP  = " << stop() << Sas::message(Msg::App,Msg::Sparse);
	    task(phkgen);
	    closeParameters();
	  }
	}
      }
    }
  }
      
  scan();

  ENDPROC
    }


void OdfObject::generateSummaryFile()  {
   PROC

     scan(); // in case we added more files

   purifySets(); // global hook to perform last minute modifications to the files we created.

    // FIXME -- this should be done differently
   string cmd = "touch " + _dir + "/" + rof();
   system(cmd.c_str());
   file_list.push_back(rof());



   file_list.push_back(osf());

   typedef vector<string>::const_iterator VI;

   string sum = _dir + "/" + osf();
   
   ofstream os;
   os.open(sum.c_str());

   errstr << "Generating the ODF summary file " + sum <<
     Sas::message(Msg::App,Msg::Sparse);
   
   observation_rec(os);
   file_details_rec(os);

   for(VI i = instruments.begin(); i != instruments.end(); i++) {
     string inst = *i;

     instrument_details_rec(inst, os);

     typedef multimap<string,string,less<string> >::const_iterator MI;
     MI lb = exposure_map_by_inst.lower_bound(inst);
     MI ub = exposure_map_by_inst.upper_bound(inst);
     
     set<string, less<string> > exposures;
     for(MI j = lb; j != ub; j++) {
       
       pair<set<string, less<string> >::const_iterator, bool> p = exposures.insert((*j).second);

       if(p.second) {
	 exposure_sub_record(inst, (*j).second, os);
	 configuration_sub_record(inst, os);
       }
       
     }
          
   }
   
   proposal_summary_rec(os);
   
   ENDPROC
}

void OdfObject::generateTimeCorrelationFile() {
  PROC
    
  openParameters("tcfgen");
  setParameter("tcf", _dir + "/" + tcf());
  setParameter("tstart", start().toString());
  setParameter("tstop", stop().toString());
  errstr << "Generating TCF " + tcf() <<
    Sas::message(Msg::App,Msg::Sparse);
  task(tcfgen);
  closeParameters();
    
  ENDPROC
    }

void OdfObject::generateAttitudeHistoryFile() {
  PROC

    if(dataSetServer->exists(_dir + "/" + ahf())) {
      errstr << "The attitude history file will not be overwritten. If this is not what you want, remove " + _dir + "/" + ahf() + " and start again." << Sas::warning("AttitudeHistoryFileExists");
    } else {
      
      openParameters("ahfgen");
      setParameter("ahf", _dir + "/" + ahf());
      setParameter("tstart", start().toString());
      setParameter("tstop", stop().toString());
      setParameter("ra", pointingRa());
      setParameter("dec", pointingDec());
      setParameter("posangle", astronomicalPositionAngle());
      errstr << "Generating AHF " + ahf() <<
	Sas::message(Msg::App,Msg::Sparse);
      task(ahfgen);
      closeParameters();
    }

  ENDPROC
}


double OdfObject::pointingRa() const {
  PROC
  RETURN _ra; // degrees
}

double OdfObject::pointingDec() const {
  PROC
  RETURN _dec; // degrees
}

int OdfObject::observationDuration() const {
  PROC
  RETURN static_cast<int>(stop() - start());
}

double OdfObject::astronomicalPositionAngle() const {
  PROC
  RETURN _posangle; // degrees
}




// ---------------
// private methods
// ---------------

void OdfObject::purifySets() const
{
  PROC
  Directory d(_dir);
  // files we might have created
  d.filter("*PFH*.FIT|*PHE*.FIT|*WDX*.FIT|*TCS*.FIT|*ATS*.FIT|*PMH*.FIT|*THX*.FIT|*NPH*.FIT");

  for(Directory::const_iterator i = d.begin(); i != d.end(); i++){
    errstr << "Purifying " << (*i)->name() << Sas::message(Msg::App,Msg::Noisy);
    DataSet * set = dataSetServer->open((*i)->name(), DataSet::Modify, DataSetServer::Low);
    if(set->hasAttribute("FILENAME")){
      string filename = set->attribute("FILENAME")->asString();
      * set->attribute("FILENAME") = filename.erase(0,filename.size()-31);
    }
    for(unsigned int j = 0; j < set->blocks(); j++){
      Block * block = set->block(j);
      if(block->hasAttribute("FILENAME")){
	string filename = block->attribute("FILENAME")->asString();
	* block->attribute("FILENAME") = filename.erase(0,filename.size()-31);
      }
    } 
    dataSetServer->close(set);
  }

  // RGS auxiliary files -- fix wrap around problem
  d.filter("*R1*AUX.FIT|*R2*AUX.FIT");
  for(Directory::const_iterator i = d.begin(); i != d.end(); i++){
    openParameters("rgsauxtable");
    setParameter("set", (*i)->name());
    task(rgsauxtable);
  }

  ENDPROC
}

void OdfObject::scan(const bool checkDate = false) {
  PROC

  Directory d(_dir);

  // This patter matches the file name structure defined in the XMM
  // ODF ICD, version 2.0
  //  d.filter("????_??????????_???????????.FIT");
  // First look only at the *.FIT files flagged as being Scheduled or
  // Unscheduled. These define the observation structure in terms of
  // exposures and active istruments.
  d.filter("????_??????????_??S????????.FIT|????_??????????_??U????????.FIT");
  typedef Directory::const_iterator LI;
  
  errstr << "Scanning ODF directory " + d.name() <<  Sas::message(Msg::App,Msg::Sparse);

  for(LI j = d.begin(); j != d.end(); j++) {
    const DirEntry &e = **j;

    // See if we have this constituent already.
    pair<set<string, less<string> >::const_iterator, bool> p =
      file_set.insert(e.name());

    if(p.second) { // this is a file we have not processed yet

      errstr << "Found " + e.basename() << Sas::message(Msg::App,Msg::Sparse);
      OdfConstituent oc(e.name());
      
      // Initializes the observation ID, the revolution number, and the
      // proposal ID. These are unique to one ODF, so we take the first
      // one we find and discard all the other files. This is not
      // optimum, but it is unlikely to be a severe limitation.
      
      if(observation_id == ""){
	observation_id = oc.observation();
	errstr << "Identified this observation: " << observation_id << Sas::message(Msg::App,Msg::Sparse);
      }
      if(revolution == ""){
	revolution = oc.revolution();
	errstr << "Identified this revolution: " << revolution << Sas::message(Msg::App,Msg::Sparse);
      }
      if(proposal == ""){
	proposal = oc.proposal();
	errstr << "Identified this proposal: " << proposal << Sas::message(Msg::App,Msg::Sparse);
      }

      // Ignore constituents that seemto belong to a different ODF.

      if(!(observation_id == oc.observation())) {
	errstr << "Observation " << observation_id << ". This component " <<
	  oc.base_name() << " is from a different ODF. Skipped" << Sas::warning("MultipleODF");
	continue;
      }
      if(!(revolution == oc.revolution())) {
	errstr << "Revolution " << revolution << ". This component " <<
	  oc.base_name() << " is from a different ODF. Skipped" << Sas::warning("MultipleODF");
	continue;
      }
      if(!(proposal == oc.proposal())) {
	errstr << "Proposal " << proposal << ". This component " <<
	  oc.base_name() << " is from a different ODF. Skipped" << Sas::warning("MultipleODF");
	continue;
      }
      
      file_list.push_back(oc.base_name());      
      pair<const string, OdfConstituent> pf(oc.instrument(),oc);
      file_map_by_inst.insert(pf);      
      pair<const string, string> pm(oc.instrument(),oc.sux_exposure());
      exposure_map_by_inst.insert(pm);
      
      // Check the date only for ODF constituents that are either
      // scheduled (S) or unscheduled (U). Constituents for whhc the
      // scheduled flag does not apply (X) refer to the whole
      // observation (DATE-OBS = earliest time in the observation,
      // DATE-END = latest time in the observation).
            
      if(oc.scheduled() != "X" && checkDate){
	DataSet * set = dataSetServer->open(oc.name(), DataSet::Read, DataSetServer::Low);
      
	// Access by number as some ODF components may not have block
	// names. Also block names encode the block number ...

	Block * blk = set->block(0); 
	ExposureObject this_exp;
	this_exp.exposure = oc.exposure();
	this_exp.scheduled = oc.scheduled();

	// All this convoluted logic is required because odffix is
	// supporting a variety of of ODF or pseudo-ODF components,
	// not all of which have the correct structure. So we try not
	// to die on a rouge file ...

	if(blk->hasAttribute("DATE-OBS")) {
	  if(isValidFitsTimeString(blk->attribute("DATE-OBS")->asString())) {
	    STime this_exp_dstart(blk->attribute("DATE-OBS")->asString());
	    if(this_exp_dstart < start_time) start_time =
					       this_exp_dstart;
	    this_exp.start = this_exp_dstart;
	    errstr << oc.base_name() << " DATE-OBS = " << this_exp.start << Sas::message(Msg::App, Msg::Verbose);
	    _has_good_start=true;
	  } else {
	    errstr << oc.base_name() << " has an invalid DATE-OBS " << blk->attribute("DATE-OBS")->asString() << " . This dataset will not be used to determine the observation duration." << Sas::warning("invalidDateString");
	  }
	} else {
	  errstr << oc.base_name() << " does not have DATE-OBS. This dataset will not be used to determine the observation duration." << Sas::warning("missingAttribute");
	  dataSetServer->close(set);
	  continue;
	} // DATE-OBS
	
	if(blk->hasAttribute("DATE-END")) {
	  if(isValidFitsTimeString(blk->attribute("DATE-END")->asString())) {
	    STime this_exp_dend(blk->attribute("DATE-END")->asString());
	    if(this_exp_dend > end_time) end_time = this_exp_dend;
	    this_exp.stop = this_exp_dend;
	    errstr << oc.base_name() << " DATE-END = " << this_exp.stop << Sas::message(Msg::App, Msg::Verbose);
	    _has_good_end=true;
	  } else {
	    errstr << oc.base_name() << " has an invalid DATE-END " << blk->attribute("DATE-END")->asString() << " . This dataset will not be used to determine the observation duration." << Sas::warning("invalidDateString");
	  }
	} else {
	  errstr << oc.base_name() << " does not have DATE-END. This dataset will not be used to determine the observation duration." <<
	    Sas::warning("missingAttribute");
	  dataSetServer->close(set);
	  continue;
	}// DATE-END
	dataSetServer->close(set);
	pair<const string, ExposureObject> pe(oc.instrument(), this_exp);
	inst_exp_time_map.insert(pe);
      } // !X and checkDate
      
    } else {
      errstr << "I already have " + e.basename() <<
	Sas::message(Msg::App,Msg::Noisy);
    } // I have this file already?
  } // loop over all directory entries

  // Count up also the ascii files and the FITS files with X-scheduled flag.
  d.filter("????_??????????_??X????????.FIT");
  for(LI j = d.begin(); j != d.end(); j++) {
    const DirEntry &e = **j;
    // See if we have this constituent already.
    pair<set<string, less<string> >::const_iterator, bool> p =
      file_set.insert(e.name());	
    if(p.second) {// this is a file we have not processed yet
	  file_list.push_back(e.basename());      
	  errstr << "Found " + e.basename() << Sas::message(Msg::App,Msg::Sparse);
	} else {
	  errstr << "I already have " + e.basename() <<
		Sas::message(Msg::App,Msg::Noisy);
	}
  }

  ENDPROC
}

// validate - 
//	+ DATE-END DATE-OBS and duration	
void OdfObject::validate() {
  PROC

  if(has_good_start()&&has_good_end()) // we have at least one usable DATE-XXX pair
    _has_good_duration = true;
  
  if(!has_good_duration()){
    errstr << "Could not find any meaningful DATE-XXX attributes to calculate the duration of the observation. Using tstart=now, tstop=now+3600s." << Sas::warning("DefaultDuration");
    if(!has_good_start()){ // no good DATE-OBS: use now for start, and reset end to be now+1h
      start_time = STime("now");
      end_time = start_time + 3600.;
    }
    if(!has_good_end()) // no good DATE-END: use end = start + 1h
      end_time = start_time + 3600.;
  }

  if(numberOfFiles() > 0){ // only check of we have ODF constituents
    if(observationDuration() <= 0)
      errstr << "Duration = " << observationDuration() <<
	". Something may be wrong with the DATE-XXX attributes." <<
	Sas::warning("NegativeDuration");
    
    if(observationDuration() > 48*86400)
      errstr << "The observation duration exceeds 48h. Something may be wrong with the DATE-XXX attributes." << Sas::warning("LongObservation");
  }

  ENDPROC
}

// file name generation methods

string OdfObject::phf(string i, string expno) const {
  PROC
  string t = "PE";
  if(i == "R1" || i == "R2")
    t = "PF"; // RGS have the Full Periodic Housekeeping instead
  string s = revolution + "_" + proposal + "_" + i +"X" + expno + "00" + t + "H.FIT";
  RETURN s;
}

string OdfObject::mphf(string i, string expno) const {
  PROC
  string s = revolution + "_" + proposal + "_" + i +"X" + expno + "00PMH.FIT";
  RETURN s;
}

string OdfObject::nphf(string i, string expno) const {
  PROC
  string s = revolution + "_" + proposal + "_" + i +"X" + expno + "00NPH.FIT";
  RETURN s;
}

string OdfObject::aphf(string i, string expno) const {
  PROC
  string s = revolution + "_" + proposal + "_" + i +"X" + expno + "00PAH.FIT";
  RETURN s;
}

string OdfObject::pwdf(string i, string expno) const {
  PROC
  string s = revolution + "_" + proposal + "_" + i + "S" + expno + "00WDX.FIT";
  RETURN s;
}

string OdfObject::thdf(string i, string expno) const {
  PROC
  string s = revolution + "_" + proposal + "_" + i + "S" + expno + "00THX.FIT";
  RETURN s;
}

string OdfObject::rof() const {
  PROC
  string s = revolution + "_" + proposal + "_SCX00000ROS.ASC";
  RETURN s;
}

string OdfObject::tcf() const {
  PROC
  string s = revolution + "_" + proposal + "_SCX00000TCS.FIT";
  RETURN s;
}

string OdfObject::osf() const {
  PROC
  string s = revolution + "_" + proposal + "_SCX00000SUM.ASC";
  RETURN s;
}

string OdfObject::ahf() const {
  PROC
  string s = revolution + "_" + proposal + "_SCX00000ATS.FIT";
  RETURN s;
}

void OdfObject::observation_rec(ofstream & os) const {
  PROC
  os << setw(11) << "OBSERVATION" << endl;
  os << setw(10) << observation_id << endl;
  os << setw(4)  << revolution << endl;
  os << setw(20) << start_time << endl;
  os << setw(20) << end_time << endl;
  ENDPROC
}

void OdfObject::file_details_rec(ofstream & os) const {
  PROC
  os << setw(5) << "FILES" << endl;
  os << setw(3) << file_list.size() << endl;
  typedef vector<string>::const_iterator VI;
  for(VI i = file_list.begin(); i != file_list.end(); i++) {
    const string &s = *i;
    os << setw(31) << s << endl;
  }
  ENDPROC
}

bool OdfObject::is_instrument_active(string inst) const {
  PROC
  bool a = false;
  if(number_of_exposures_of_instrument(inst) > 0) a = true;
  RETURN a;
}

unsigned int OdfObject::number_of_exposures_of_instrument(string inst) const {
  PROC

      // Here equal_range should work, but I cannot make it work.
    typedef multimap<string,OdfConstituent,less<string> >::const_iterator MI;
  MI lb = file_map_by_inst.lower_bound(inst);
  MI ub = file_map_by_inst.upper_bound(inst);
  
  set<string,less<string> > s;
  
  for(MI i = lb; i != ub; i++) {
    s.insert((*i).second.sux_exposure());
  }


  RETURN s.size();
}

void OdfObject::instrument_details_rec(string inst, ofstream & os) const {
  PROC
    

    unsigned int n = number_of_exposures_of_instrument(inst);

  string active = "Y";
  if(n == 0) active = "N";
    
  os << setw(10) << "INSTRUMENT" << endl;
  os << setw(2) << inst << endl;
  os << setw(1) << active << endl;
  os << setw(4) << n << endl;

  ENDPROC
}

void OdfObject::exposure_sub_record(string inst, string exp, ofstream & os) const {
  PROC

  typedef multimap<string,ExposureObject,less<string> >::const_iterator EI;
  EI lb = inst_exp_time_map.lower_bound(inst);
  EI ub = inst_exp_time_map.upper_bound(inst);
  STime exp_dstart = STime::latest;
  STime exp_dend = STime::earliest;
  // FIXME: another contorsion
  bool scheduled = (exp[0] == 'S');
  errstr << "Instrument = " << inst << Sas::message(Msg::App, Msg::Verbose);
  for(EI i = lb; i != ub; i++) {
    const  ExposureObject &e = (*i).second;
    // The exposure string is now '?nnn', where ? is one of 'S' 'U',
    // so we have to combine two strings from the ExposureObject,
    // pending a better design. FIXME.    
    if(e.scheduled + e.exposure == exp) {
      if(e.start < exp_dstart) exp_dstart = e.start;
      if(e.stop > exp_dend) exp_dend = e.stop;
      errstr << exp << " " << e.start << " " << e.stop << Sas::message(Msg::App, Msg::Verbose);
    }
  }
  
  string commanded_exp_id = "??????????";
  string proposal_exp_id = proposal + inst + exp;

  os << setw(8) << "EXPOSURE" << endl;
  os << setw(4) <<  exp.substr(1,4) << endl;
  os << setw(1) << (scheduled ? "Y" : "N") << endl;
  os << setw(7) << "SCIENCE" << endl;
  os << setw(10) << commanded_exp_id << endl;
  os << setw(15) << proposal_exp_id << endl;
  os << setw(19) << exp_dstart << endl;
  os << setw(19) << exp_dend << endl;
  os << setw(19) << exp_dstart << endl;
  os << setw(19) << exp_dend << endl;
	    
  ENDPROC
}

void OdfObject::configuration_sub_record(string inst, ofstream & os) {
  PROC

    os << setw(22) << "CONFIGURATION " + long_names[inst] << endl;
  os << setw(4) << 0 << endl;
  
  ENDPROC
}

void OdfObject::proposal_summary_rec(ofstream & os) const {
  PROC

  os << setw(8) << "PROPOSAL" << endl;
  os << setw(5) << "Dr. Pino Paperino" << endl;
  os << setw(30) << "Duck Research" << endl;
  os << setw(30) << "Duck Laboratories Unlimited" << endl;
  os << setw(30) << "Via degli Stagni" << endl;
  os << setw(20) << "Paperopoli" << endl;
  os << "State" << endl << "Country" << endl;
  os << setw(10) << "DUCK 42" << endl;
  os << setw(80) << "paperino@quack.duck-research.com" << endl;
  os << setw(2) << "AO" << endl;
  os << setw(2) << "ST" << endl;
  os << setw(20) << "duckling" << endl;
  os << setw(10) << setprecision(15) << pointingRa()/15. << endl;
  os << setw(11) << pointingDec() << endl;
  os << setw(6) << observationDuration() << endl;
  os << setw(80) << "paperini" << endl;
  os << setw(10) << pointingRa()/15. << endl;
  os << setw(11) << pointingDec() << endl;
  os << setw(3) << "0" << endl;
  os << setw(3) << "360" << endl;
  os << setw(1) << "?" << endl;
  os << setw(1) << "1" << endl;
  os << setw(1) << "2" << endl;
  os << setw(1) << "3" << endl;
  os << setw(1) << "4" << endl;
  os << setw(1) << "5" << endl;
  os << setw(1) << "6" << endl;
 
  ENDPROC
}

bool OdfObject::has_good_duration() const
{
  PROC
  RETURN _has_good_duration;
}
bool OdfObject::has_good_start() const
{
  PROC
  RETURN _has_good_start;
}
bool OdfObject::has_good_end() const
{
  PROC
  RETURN _has_good_end;
}

void OdfObject::report() const
{
  PROC
  errstr << "odffix found the following ODF: " << endl;
  errstr << "------------------------------------------------------" << endl;
  errstr << "Observation identifier : " << observation_id << endl;
  errstr << "Revolution identifier  : " << revolution << endl;
  errstr << "Proposal identifier    : " << proposal << endl;
  errstr << "Start time             : " << start() << endl;
  errstr << "Stop time              : " << stop() << endl;
  errstr << "Duration               : " << observationDuration() << endl;
  errstr << "Right Ascension        : " << pointingRa() << endl;
  errstr << "Declination            : " << pointingDec() << endl;
  errstr << "Astronomical Pos. Angle: " << astronomicalPositionAngle() << endl;
  errstr << "------------------------------------------------------" << endl;
  errstr << Sas::message(Msg::App,Msg::Sparse);
  ENDPROC
}
