//
// XMM Science Analysis System
//
// Giuseppe Vacanti (European Space Agency) 
// September 1999
//
// 
#ifndef OdfObject_H
#define OdfObject_H

#include <string>
#include <list>

#include <vector>
#include <multimap.h>
#include <map>
#include <set>
#include <fstream.h>

#include "Directory.h"
#include "OdfConstituent.h"
#include "ExposureObject.h"
#include "STime.h"

class OdfObject {
  
 public:

  OdfObject(double ra, double dec, double posangle, string dir = ".");
  OdfObject(string dir = ".");
  
  unsigned int numberOfFiles() const;

  STime start() const;
  STime stop() const;

  void generateSummaryFile();
  void generateTimeCorrelationFile();
  void generateAttitudeHistoryFile();
  void generateHousekeepingFiles();
  void generateAuxiliaryFiles();

  double pointingRa() const;
  double pointingDec() const;
  double astronomicalPositionAngle() const;
  int observationDuration() const;

 private:
  
  void init();
  void purifySets() const;
  void scan(const bool checkDate = false);
  void observation_rec(ofstream & os) const;
  void file_details_rec(ofstream & os) const;
  void instrument_details_rec(string inst, ofstream & os) const;
  void exposure_sub_record(string inst, string exp, ofstream & os) const;
  void configuration_sub_record(string inst, ofstream & os);
  void proposal_summary_rec(ofstream & os) const;
  bool has_good_duration() const;
  bool has_good_start() const;
  bool has_good_end() const;
  void validate();
  void report() const;

  string osf() const; // ODF summary file
  string ahf() const; // attitude history file
  string tcf() const; // time correlation file
  string rof() const; // reconstructed orbit file (ASCII)
  
  string phf(string inst, string expno) const; // periodic housekeeping file
  string mphf(string inst, string expno) const; // main periodic housekeeping file
  string nphf(string inst, string expno) const; // non periodic housekeeping file
  string aphf(string inst, string expno) const; // auxiliary periodic housekeeping file
  string pwdf(string inst, string expno) const; // priority window data auxiliary file
  string thdf(string inst, string expno) const; // tracking history data auxiliary file

  bool is_instrument_active(string inst) const;
  unsigned int number_of_exposures_of_instrument(string inst) const;

  multimap<string,OdfConstituent,less<string> > file_map_by_inst;
  multimap<string,string,less<string> > exposure_map_by_inst; 
  multimap<string,ExposureObject,less<string> > inst_exp_time_map;

  vector<string> instruments;
  vector<string> file_list;

  map<string,string,less<string> > long_names;
  map<string,string,less<string> > fits_names;
  
  set<string, less<string> > file_set;
  
  string revolution;
  string proposal;

  string observation_id;
  STime start_time;
  STime end_time;

  string _dir;

  bool _pointing_from_ats; // read RA DEC POSANGLE from attitude history file?

  double _ra, _dec, _posangle;

  bool _has_good_duration, _has_good_start, _has_good_end;

};

#endif
